#include "trucs_sockets.h"
#include <iostream>
using std::cerr;
using std::cout;
using std::endl;
#include <string>
using std::string;
#include <Windows.h>

class Mutex
{
   HANDLE mutex_;
public:
   Mutex() throw()
      : mutex_(CreateMutex(0, FALSE, 0))
   {
   }
   ~Mutex() throw()
      { CloseHandle(mutex_); }
   void obtenir() const volatile throw()
      { WaitForSingleObject(mutex_, INFINITE); }
   void relacher() const volatile throw()
      { ReleaseMutex(mutex_); }
};

class Autoverrou
{
   const volatile Mutex &m_;
public:
   Autoverrou(const volatile Mutex &m)
      : m_(m)
   {
      m_.obtenir();
   }
   ~Autoverrou() throw()
      { m_.relacher(); }
};

class zone_transit
{
   string texte_;
   Mutex mutex_;
public:
   void ajouter(const string &s) volatile
   {
      Autoverrou av(mutex_);
      const_cast<zone_transit&>(*this).ajouter(s);
   }
   void ajouter(const string &s)
      { texte_ += s; }
   string extraire() volatile
   {
      Autoverrou av(mutex_);
      return const_cast<zone_transit&>(*this).extraire();
   }
   string extraire()
   {
      string s = texte_;
      texte_.clear();
      return s;
   }
};

struct Relais
{
   SOCKET sck;
   zone_transit sortie;
   zone_transit entree;
   Relais(SOCKET sck)
      : sck(sck), fini(false)
   {
   }
private:
   bool fini;
public:
   bool est_fini() const
      { return fini; }
   void fermer()
   {
      // condition de course!
      if (!fini)
      {
         fini = true;
         FermerSocket(sck);
      }
   }
   ~Relais() throw()
      { fermer(); }
};

#include "autonome.h"
#include <memory>
using std::unique_ptr;

class LireClavier
   : public Autonome
{
   Relais &r;
   LireClavier(Relais &r)
      : Autonome(0), r(r)
   {
   }
   void agir()
   {
      using std::cin;
      using std::getline;
      string s;
      if (!getline(cin, s) || s == ".")
      {
         r.fermer();
         meurs();
      }
      else
         r.sortie.ajouter(s + '\n');
   }
public:
   static unique_ptr<Autonome> creer(Relais &r)
   {
      LireClavier *p = new LireClavier(r);
      p->demarrer();
      return unique_ptr<Autonome>(p);
   }
};

class EnvoyerSocket
   : public Autonome
{
   Relais &r;
   EnvoyerSocket(Relais &r)
      : Autonome(0), r(r)
   {
   }
   void agir()
   {
      if (r.est_fini())
         meurs();
      try
      {
         string s = r.sortie.extraire();
         if (!s.empty())
            Envoyer(r.sck, s, '\0');
      }
      catch(erreur &e)
      {
         cerr << e << endl;
         r.fermer();
         meurs();
      }
   }
public:
   static unique_ptr<Autonome> creer(Relais &r)
   {
      EnvoyerSocket *p = new EnvoyerSocket(r);
      p->demarrer();
      return unique_ptr<Autonome>(p);
   }
};

class ConsommerSocket
   : public Autonome
{
   Relais &r;
   ConsommerSocket(Relais &r)
      : Autonome(0), r(r)
   {
   }
   void agir()
   {
      try
      {
         string s = Recevoir(r.sck, '\0');
         if (!s.empty())
            r.entree.ajouter(s);
      }
      catch(erreur &e)
      {
         cerr << e << endl;
         r.fermer();
         meurs();
      }
   }
public:
   static unique_ptr<Autonome> creer(Relais &r)
   {
      ConsommerSocket *p = new ConsommerSocket(r);
      p->demarrer();
      return unique_ptr<Autonome>(p);
   }
};

class EcrireConsole
   : public Autonome
{
   Relais &r;
   EcrireConsole(Relais &r)
      : Autonome(0), r(r)
   {
   }
   void agir()
   {
      using std::flush;
      if (r.est_fini())
         meurs();
      try
      {
         string s = r.entree.extraire();
         if (!s.empty())
            cout << s << flush;
      }
      catch(erreur &e)
      {
         cerr << e << endl;
         r.fermer();
         meurs();
      }
   }
public:
   static unique_ptr<Autonome> creer(Relais &r)
   {
      EcrireConsole *p = new EcrireConsole(r);
      p->demarrer();
      return unique_ptr<Autonome>(p);
   }
};

#include <vector>
using std::vector;

int main()
{
   const short PORT_A_RESERVER= normaliser(static_cast<short>(4321));
   const string MESSAGE = "Reu 5 sur 5";
   const char DELIMITEUR_FIN = '#';
   // Initialisation des sockets pour le serveur
   ChargerSockets();
   // Obtention du socket serveur
   socket_t serveur = CreerSocket();
   // On rend le serveur disponible en rservant son port
   ReserverPort(serveur, PORT_A_RESERVER);
   // Acceptation du client
   socket_t pourJaser = AccepterClient(serveur);

   bool fini = false;
   Relais r(pourJaser);
   auto lire_clav = LireClavier::creer(r);
   auto env_sck = EnvoyerSocket::creer(r);
   auto lire_sck = ConsommerSocket::creer(r);
   auto ecr_cons = EcrireConsole::creer(r);
   vector<unique_ptr_ref<Autonome>> v;
   v.push_back(lire_clav);
   v.push_back(env_sck);
   v.push_back(lire_sck);
   v.push_back(ecr_cons);
   attendre(v);
   FermerSocket(serveur);
   DechargerSockets();
}
