#include "autonome.h"
#include <Windows.h>

void Rythmique::dormir() const
   { Sleep(dodo_); }

namespace
{
   unsigned long __stdcall ze_thread(void *p)
   {
      Autonome *a = reinterpret_cast<Autonome*>(p);
      a->debut();
      while (!a->dois_mourir())
      {
         a->agir();
         a->dormir();
      }
      a->fin();
      a->je_trepasse();
      return 0;
   }
}

#include <cassert>

struct Evenement::Rep
{
   HANDLE h;
   Rep(HANDLE h)
      : h(h)
   {
      assert(h);
   }
   ~Rep() throw()
      { CloseHandle(h); }
};

#include <memory>
using std::unique_ptr;
using std::weak_ptr;

unique_ptr<Evenement::Rep> Evenement::creer(ManualReset)
{
   HANDLE h = CreateEvent(0, TRUE, FALSE, 0);
   try
      { return unique_ptr<Rep>(new Rep(h)); }
   catch(...)
   {
      CloseHandle(h);
      throw;
   }
}
unique_ptr<Evenement::Rep> Evenement::creer(AutomaticReset)
{
   HANDLE h = CreateEvent(0, FALSE, FALSE, 0);
   try
      { return unique_ptr<Rep>(new Rep(h)); }
   catch(...)
   {
      CloseHandle(h);
      throw;
   }
}

Evenement::Evenement(ManualReset reset_type)
   : p_(creer(reset_type))
{
}
Evenement::Evenement(AutomaticReset reset_type)
   : p_(creer(reset_type))
{
}

void Evenement::attendre() const volatile
   { WaitForSingleObject(const_cast<unique_ptr<Rep>&>(p_)->h, INFINITE); }
void Evenement::provoquer() const volatile
   { SetEvent(const_cast<unique_ptr<Rep>&>(p_)->h); }
void Evenement::reinitialiser() const volatile
   { ResetEvent(const_cast<unique_ptr<Rep>&>(p_)->h); }
Evenement::~Evenement() throw()
   { }

struct Autonome::Rep
{
   HANDLE h;
   Rep(HANDLE h)
      : h(h)
   {
      assert(h);
   }
   ~Rep() throw()
      { CloseHandle(h); }
};

void Autonome::arreter()
{
   meurs();
   attendre_fin();
}

void Autonome::demarrer()
{
   HANDLE h = CreateThread(0, 0, ::ze_thread, this, 0, 0);
   try
   {
      p_ = unique_ptr<Rep>(new Rep(h));
   }
   catch(...)
   {
      arreter();
      CloseHandle(h);
      throw;
   }
}

Autonome::Autonome(const value_type dodo)
   : Rythmique(dodo)
{
}

Autonome::~Autonome() throw()
   { }

#include <vector>
using std::vector;
#include <algorithm>
using std::distance;
using std::for_each;
using std::begin;
using std::end;

void attendre(const vector<unique_ptr_ref<Autonome>> &v)
{
   vector<HANDLE> h;
   for_each(begin(v), end(v), [&h](const unique_ptr_ref<Autonome> &p) {
      h.push_back(p.ref->p_->h);
   });
   WaitForMultipleObjects(v.size(), &h[0], TRUE, INFINITE);
}
