#ifndef AUTONOME_H
#define AUTONOME_H

struct Acteur
{
   virtual void agir() = 0;
   virtual ~Acteur() throw()
      { }
};

class Rythmique
{
public:
   typedef long value_type;
private:
   value_type dodo_;
public:
   Rythmique(const value_type dodo)
      : dodo_(dodo)
   {
   }
   void dormir() const;
};

#include <memory>

class Evenement
{
   struct Rep;
   std::unique_ptr<Rep> p_;
public:
   class ManualReset{};
   class AutomaticReset{};
private:
   static std::unique_ptr<Rep> creer(ManualReset);
   static std::unique_ptr<Rep> creer(AutomaticReset);
public:
   Evenement(ManualReset);
   Evenement(AutomaticReset);
   ~Evenement() throw();
   void attendre() const volatile;
   void provoquer() const volatile;
   void reinitialiser() const volatile;
};

class Stoppable
{
   bool dois_mourir_,
        est_mort_;
   Evenement deces_;
public:
   Stoppable() throw()
      : dois_mourir_(false), est_mort_(false), deces_(Evenement::ManualReset())
   {
   }
   void meurs()
      { dois_mourir_ = true; }
   void je_trepasse()
   {
      est_mort_ = true;
      deces_.provoquer();
   }
   bool est_trepasse() const
      { return est_mort_; }
   bool dois_mourir() const
      { return dois_mourir_; }
   void attendre_fin() const
      { deces_.attendre(); }
};

template <class T>
   struct unique_ptr_ref
   {
      std::unique_ptr<T> &ref;
      unique_ptr_ref(std::unique_ptr<T> &ref)
         : ref(ref)
      {
      }
   };

#include <vector>

class Autonome
   : public Acteur, public Rythmique, public Stoppable
{
   struct Rep;
   std::unique_ptr<Rep> p_;
   void arreter();
protected:
   Autonome(const value_type dodo = value_type());
   void demarrer();
public:
   ~Autonome() throw();
   virtual void debut() {}
   virtual void fin() {}
   friend void attendre(const std::vector<unique_ptr_ref<Autonome>>&);
};

#endif
